#!/usr/bin/env python3
import base64
import datetime
import time
import uuid
import decimal
import bson
import singer
from singer import utils, metadata
from bson import objectid, timestamp, encode, decode, datetime as bson_datetime
from bson.binary import Binary, UuidRepresentation
from bson.codec_options import CodecOptions, DatetimeConversion
from bson.datetime_ms import DatetimeMS

import pytz
import tzlocal

INCLUDE_SCHEMAS_IN_DESTINATION_STREAM_NAME = False
UPDATE_BOOKMARK_PERIOD = 1000
COUNTS = {}
TIMES = {}
SCHEMA_COUNT = {}
SCHEMA_TIMES = {}

LOGGER = singer.get_logger()

class InvalidProjectionException(Exception):
    """Raised if projection blacklists _id"""

class UnsupportedReplicationKeyTypeException(Exception):
    """Raised if key type is unsupported"""

class MongoAssertionException(Exception):
    """Raised if Mongo exhibits incorrect behavior"""

class MongoInvalidDateTimeException(Exception):
    """Raised if we find an invalid date-time that we can't handle"""

def calculate_destination_stream_name(stream):
    s_md = metadata.to_map(stream['metadata'])
    if INCLUDE_SCHEMAS_IN_DESTINATION_STREAM_NAME:
        return "{}_{}".format(s_md.get((), {}).get('database-name'), stream['stream'])

    return stream['stream']

def whitelist_bookmark_keys(bookmark_key_set, tap_stream_id, state):
    for bookmark_key in [non_whitelisted_bookmark_key
                         for non_whitelisted_bookmark_key
                         in state.get('bookmarks', {}).get(tap_stream_id, {}).keys()
                         if non_whitelisted_bookmark_key not in bookmark_key_set]:
        singer.clear_bookmark(state, tap_stream_id, bookmark_key)


def get_stream_version(tap_stream_id, state):
    stream_version = singer.get_bookmark(state, tap_stream_id, 'version')

    if stream_version is None:
        stream_version = int(time.time() * 1000)

    return stream_version

def class_to_string(bookmark_value, bookmark_type):
    if bookmark_type == 'datetime':
        timezone = tzlocal.get_localzone()
        local_datetime = timezone.localize(bookmark_value)
        utc_datetime = local_datetime.astimezone(pytz.UTC)
        return utils.strftime(utc_datetime)
    if bookmark_type == 'Timestamp':
        return '{}.{}'.format(bookmark_value.time, bookmark_value.inc)
    if bookmark_type == 'bytes':
        return base64.b64encode(bookmark_value).decode('utf-8')
    if bookmark_type in ['int', 'Int64', 'float', 'ObjectId', 'str', 'UUID']:
        return str(bookmark_value)
    if bookmark_type == 'Binary':
        return str(bookmark_value.as_uuid(bookmark_value.subtype))

    raise UnsupportedReplicationKeyTypeException("{} is not a supported replication key type"
                                                 .format(bookmark_type))


# pylint: disable=too-many-return-statements
def string_to_class(str_value, type_value):
    if type_value == 'UUID':
        return uuid.UUID(str_value)
    if type_value == 'datetime':
        return singer.utils.strptime_with_tz(str_value)
    if type_value == 'int':
        return int(str_value)
    if type_value == 'Int64':
        return bson.int64.Int64(str_value)
    if type_value == 'float':
        return float(str_value)
    if type_value == 'ObjectId':
        return objectid.ObjectId(str_value)
    if type_value == 'Timestamp':
        split_value = str_value.split('.')
        return bson.timestamp.Timestamp(int(split_value[0]), int(split_value[1]))
    if type_value == 'bytes':
        return base64.b64decode(str_value.encode())
    if type_value == 'str':
        return str(str_value)
    if type_value == 'Binary':
        return Binary.from_uuid(uuid.UUID(str_value), UuidRepresentation.STANDARD)
    raise UnsupportedReplicationKeyTypeException("{} is not a supported replication key type"
                                                 .format(type_value))

def safe_transform_datetime(value, path):
    timezone = tzlocal.get_localzone()
    try:
        local_datetime = timezone.localize(value)
        utc_datetime = local_datetime.astimezone(pytz.UTC)
    except Exception as ex:
        if str(ex) == "year is out of range" and value.year == 0:
            # NB: Since datetimes are persisted as strings, it doesn't
            # make sense to blow up on invalid Python datetimes (e.g.,
            # year=0). In this case we're formatting it as a string and
            # passing it along down the pipeline.
            return "{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:02d}.{:06d}Z".format(value.year,
                                                                              value.month,
                                                                              value.day,
                                                                              value.hour,
                                                                              value.minute,
                                                                              value.second,
                                                                              value.microsecond)
        raise MongoInvalidDateTimeException("Found invalid datetime at [{}]: {}".format(
            ".".join(map(str, path)),
            value)) from ex
    return utils.strftime(utc_datetime)

# pylint: disable=too-many-return-statements,too-many-branches
def transform_value(value, path):
    if isinstance(value, list):
        # pylint: disable=unnecessary-lambda
        return list(map(lambda v: transform_value(v[1], path + [v[0]]), enumerate(value)))
    if isinstance(value, dict):
        return {k:transform_value(v, path + [k]) for k, v in value.items()}
    if isinstance(value, uuid.UUID):
        return str(value)
    if isinstance(value, objectid.ObjectId):
        return str(value)
    if isinstance(value, bson_datetime.datetime):
        return safe_transform_datetime(value, path)
    if isinstance(value, timestamp.Timestamp):
        return utils.strftime(value.as_datetime())
    if isinstance(value, bson.int64.Int64):
        return int(value)
    if isinstance(value, Binary):
        return str(value.as_uuid(value.subtype))
    if isinstance(value, bytes):
        # Return the original base64 encoded string
        return base64.b64encode(value).decode('utf-8')
    if isinstance(value, datetime.datetime):
        timezone = tzlocal.get_localzone()
        local_datetime = timezone.localize(value)
        utc_datetime = local_datetime.astimezone(pytz.UTC)
        return utils.strftime(utc_datetime)
    if isinstance(value, bson.decimal128.Decimal128):
        return value.to_decimal()
    if isinstance(value, bson.code.Code):
        if value.scope:
            return {
                'value': str(value),
                'scope': str(value.scope)
            }
        return str(value)
    if isinstance(value, bson.regex.Regex):
        return {
            'pattern': value.pattern,
            'flags': value.flags
        }
    if isinstance(value, bson.dbref.DBRef):
        return {
            'id': str(value.id),
            'collection': value.collection,
            'database': value.database
        }
    if isinstance(value, DatetimeMS):
        encoded_datetime = encode({"datetime_value": value})
        codec_option = CodecOptions(datetime_conversion=DatetimeConversion.DATETIME_CLAMP)
        decoded_datetime = decode(encoded_datetime, codec_options=codec_option)
        return safe_transform_datetime(decoded_datetime["datetime_value"], path)

    return value

def row_to_singer_record(stream, row, version, time_extracted):
    # pylint: disable=unidiomatic-typecheck
    try:
        row_to_persist = {k:transform_value(v, [k]) for k, v in row.items()
                          if type(v) not in [bson.min_key.MinKey, bson.max_key.MaxKey]}
    except MongoInvalidDateTimeException as ex:
        raise Exception("Error syncing collection {}, object ID {} - {}".format(stream["tap_stream_id"], row['_id'], ex)) from ex

    return singer.RecordMessage(
        stream=calculate_destination_stream_name(stream),
        record=row_to_persist,
        version=version,
        time_extracted=time_extracted)

def add_to_any_of(schema, value):
    changed = False

    if isinstance(value, (bson_datetime.datetime, timestamp.Timestamp, datetime.datetime)):
        has_date = False
        for field_schema_entry in schema:
            if field_schema_entry.get('format') == 'date-time':
                has_date = True
                break
        if not has_date:
            schema.insert(0, {"type": "string", "format": "date-time"})
            changed = True

    elif isinstance(value, bson.decimal128.Decimal128):
        has_date = False
        has_decimal = False

        for field_schema_entry in schema:
            if field_schema_entry.get('format') == 'date-time':
                has_date = True
            if field_schema_entry.get('type') == 'number' and not field_schema_entry.get('multipleOf'):
                field_schema_entry['multipleOf'] = decimal.Decimal('1e-34')
                return True
            if field_schema_entry.get('type') == 'number' and field_schema_entry.get('multipleOf'):
                has_decimal = True

        if not has_decimal:
            if has_date:
                schema.insert(1, {"type": "number", "multipleOf": decimal.Decimal('1e-34')})
            else:
                schema.insert(0, {"type": "number", "multipleOf": decimal.Decimal('1e-34')})
            changed = True

    elif isinstance(value, float):
        has_date = False
        has_float = False

        for field_schema_entry in schema:
            if field_schema_entry.get('format') == 'date-time':
                has_date = True
            if field_schema_entry.get('type') == 'number' and field_schema_entry.get('multipleOf'):
                field_schema_entry.pop('multipleOf')
                return True
            if field_schema_entry.get('type') == 'number' and not field_schema_entry.get('multipleOf'):
                has_float = True

        if not has_float:
            if has_date:
                schema.insert(1, {"type": "number"})
            else:
                schema.insert(0, {"type": "number"})

            changed = True

    elif isinstance(value, dict):
        has_object = False

        # get pointer to object schema and see if it already existed
        object_schema = {"type": "object", "properties": {}}
        for field_schema_entry in schema:
            if field_schema_entry.get('type') == 'object':
                object_schema = field_schema_entry
                has_object = True

        # see if object schema changed
        if row_to_schema(object_schema, value):
            changed = True

            # if it changed and existed, it's reference was modified
            # if it changed and didn't exist, insert it
            if not has_object:
                schema.insert(-1, object_schema)
    elif isinstance(value, list):
        has_list = False

        # get pointer to list's anyOf schema and see if list schema already existed
        list_schema = {"type": "array", "items": {"anyOf": [{}]}}
        for field_schema_entry in schema:
            if field_schema_entry.get('type') == 'array':
                list_schema = field_schema_entry
                has_list = True
        anyof_schema = list_schema['items']['anyOf']

        # see if list schema changed
        list_entry_changed = False
        for list_entry in value:
            list_entry_changed = add_to_any_of(anyof_schema, list_entry) or list_entry_changed
            changed = changed or list_entry_changed

        # if it changed and existed, it's reference was modified
        # if it changed and didn't exist, insert it
        if not has_list and list_entry_changed:
            schema.insert(-1, list_schema)
    return changed

def row_to_schema(schema, row):
    changed = False

    for field, value in row.items():
        if isinstance(value, (bson_datetime.datetime,
                              timestamp.Timestamp,
                              datetime.datetime,
                              bson.decimal128.Decimal128,
                              float,
                              dict,
                              list)):

            # get pointer to field's anyOf list
            if not schema.get('properties', {}).get(field):
                schema['properties'][field] = {'anyOf': [{}]}
            anyof_schema = schema['properties'][field]['anyOf']

            # add value's schema to anyOf list
            changed = add_to_any_of(anyof_schema, value) or changed

    return changed

def get_sync_summary(catalog):
    headers = [['database',
                'collection',
                'replication method',
                'total records',
                'write speed',
                'total time',
                'schemas written',
                'schema build duration',
                'percent building schemas']]

    rows = []
    for stream_id, stream_count in COUNTS.items():
        stream = [x for x in catalog['streams'] if x['tap_stream_id'] == stream_id][0]
        collection_name = stream.get("table_name")
        md_map = metadata.to_map(stream['metadata'])
        db_name = metadata.get(md_map, (), 'database-name')
        replication_method = metadata.get(md_map, (), 'replication-method')

        stream_time = TIMES[stream_id]
        schemas_written = SCHEMA_COUNT[stream_id]
        schema_duration = SCHEMA_TIMES[stream_id]
        if stream_time == 0:
            stream_time = 0.000001
        row = [
            db_name,
            collection_name,
            replication_method,
            '{} records'.format(stream_count),
            '{:.1f} records/second'.format(stream_count/stream_time),
            '{:.5f} seconds'.format(stream_time),
            '{} schemas'.format(schemas_written),
            '{:.5f} seconds'.format(schema_duration),
            '{:.2f}%'.format(100*schema_duration/stream_time)
        ]
        rows.append(row)
    LOGGER.info("\n**** Sync Summary:")
    LOGGER.info(next(iter(headers), None))
    for row in rows:
        LOGGER.info(row)



import os
import random
import string
import unittest
from bson import ObjectId
from pymongo.errors import OperationFailure

from mongodb_common import drop_all_collections, get_test_connection, ensure_environment_variables_set
from tap_tester import connections, menagerie, runner, LOGGER


RECORD_COUNT = {}

# namespace length = db name + collection name + 1 (for the separator dot)
#   mongo 4.2 max supported namespace is 120 bytes = 120 characters
#   mongo 4.4, 5.0, and 6.0 max supported namespace is 255 bytes
db_name = 'simple_db'
max_name_length = 120

# collection names
coll_name_1 = f'{"Collection_name_with_110_characters_":1<110}'
coll_name_2 = f'{"Collection_name_with_110_characters_":2<110}'
coll_name_3 = f'{"Collection_name_with_111_characters_":3<111}'

coll_name_4 = f'{"Collection_name_with_245_characters_":4<245}'
coll_name_5 = f'{"Collection_name_with_245_characters_":5<245}'
coll_name_6 = f'{"Collection_name_with_246_characters_":6<246}'


def random_string_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))

def generate_simple_coll_docs(num_docs):
    docs = []
    for int_value in range(num_docs):
        docs.append({"int_field": int_value, "string_field": random_string_generator()})
    return docs

class MongoDBNameSpaceRestrictions(unittest.TestCase):
    ''' Test edge case namespace restrictions per the documentation (120 byte max)
        Reference https://jira.talendforge.org/browse/TDL-18990 for details  '''

    def setUp(self):

        ensure_environment_variables_set()

        with get_test_connection() as client:
            ############# Drop all dbs/collections #############
            drop_all_collections(client)

            db_version = client.server_info()['version']
            LOGGER.info("MongoDB version: {}".format(db_version))

            if db_version.startswith('4.2.'):  # 4.2-bionic db version string unknown BUG
                self.coll_name_1 = coll_name_1
                self.coll_name_2 = coll_name_2
                self.coll_name_3 = coll_name_3
                self.max_name_length = max_name_length
            elif db_version in ['4.4.6', '5.0.15', '6.0.4']:
                self.coll_name_1 = coll_name_4
                self.coll_name_2 = coll_name_5
                self.coll_name_3 = coll_name_6
                self.max_name_length = 255
            else:
                assert True == False, "Uknown database version detected!"

            LOGGER.info("Namespace lengths. 1: {}, 2: {}, 3: {}, Max: {}".format(
                len(self.coll_name_1) + len(db_name) + 1,
                len(self.coll_name_2) + len(db_name) + 1,
                len(self.coll_name_3) + len(db_name) + 1,
                self.max_name_length))

            ############# Add simple collections #############
            # coll_name_1 has 50 documents
            client[db_name][self.coll_name_1].insert_many(generate_simple_coll_docs(50))

            # coll_name_2 has 100 documents
            client[db_name][self.coll_name_2].insert_many(generate_simple_coll_docs(100))

            # coll_name_3 should fail
            with self.assertRaises(OperationFailure) as e:
                client[db_name][self.coll_name_3].insert_many(generate_simple_coll_docs(10))

            self.assertIn("Fully qualified namespace is too long", str(e.exception.details.values()))
            self.assertIn(f"Max: {self.max_name_length}", str(e.exception.details.values()))


    def expected_check_streams(self):
        return {
            f'{db_name}-{self.coll_name_1}',
            f'{db_name}-{self.coll_name_2}',
        }

    def expected_pks(self):
        return {
            self.coll_name_1: {'_id'},
            self.coll_name_2: {'_id'},
        }

    def expected_row_counts(self):
        return {
            self.coll_name_1: 50,
            self.coll_name_2: 100,
        }

    def expected_sync_streams(self):
        return {
            self.coll_name_1,
            self.coll_name_2
        }

    def name(self):
        return "tap_tester_mongodb_namespace_restrict"

    def tap_name(self):
        return "tap-mongodb"

    def get_type(self):
        return "platform.mongodb"

    def get_credentials(self):
        return {'password': os.getenv('TAP_MONGODB_PASSWORD')}

    def get_properties(self):
        return {'host' : os.getenv('TAP_MONGODB_HOST'),
                'port' : os.getenv('TAP_MONGODB_PORT'),
                'user' : os.getenv('TAP_MONGODB_USER'),
                'database' : os.getenv('TAP_MONGODB_DBNAME')
        }

    def test_run(self):

        conn_id = connections.ensure_connection(self)

        #  ---------------------------------
        #  -----------  Discovery ----------
        #  ---------------------------------

        # run in discovery mode
        check_job_name = runner.run_check_mode(self, conn_id)

        # verify check  exit codes
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        menagerie.verify_check_exit_status(self, exit_status, check_job_name)

        # verify the tap discovered the right streams
        found_catalogs = menagerie.get_catalogs(conn_id)

        # assert we find the correct streams
        self.assertEqual(self.expected_check_streams(),
                         {c['tap_stream_id'] for c in found_catalogs})

        for tap_stream_id in self.expected_check_streams():
            found_stream = [c for c in found_catalogs if c['tap_stream_id'] == tap_stream_id][0]

            # assert that the pks are correct
            self.assertEqual(self.expected_pks()[found_stream['stream_name']],
                             set(found_stream.get('metadata', {}).get('table-key-properties')))

            # assert that the row counts are correct
            self.assertEqual(self.expected_row_counts()[found_stream['stream_name']],
                             found_stream.get('metadata', {}).get('row-count'))

        #  ----------------------------------------
        #  ----------- Initial Full Table ---------
        #  ----------------------------------------

        # Select coll_name_1 and coll_name_2 streams and add replication method metadata
        for stream_catalog in found_catalogs:
            annotated_schema = menagerie.get_annotated_schema(conn_id, stream_catalog['stream_id'])
            additional_md = [{ "breadcrumb" : [], "metadata" : {'replication-method' : 'LOG_BASED'}}]
            selected_metadata = connections.select_catalog_and_fields_via_metadata(conn_id,
                                                                                    stream_catalog,
                                                                                    annotated_schema,
                                                                                    additional_md)

        # Run sync
        sync_job_name = runner.run_sync_mode(self, conn_id)

        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # verify the persisted schema was correct
        records_by_stream = runner.get_records_from_target_output()

        # assert that each of the streams that we synced are the ones that we expect to see
        record_count_by_stream = runner.examine_target_output_file(self,
                                                                   conn_id,
                                                                   self.expected_sync_streams(),
                                                                   self.expected_pks())

        # Verify that the full table was synced
        for tap_stream_id in self.expected_sync_streams():
            self.assertGreaterEqual(record_count_by_stream[tap_stream_id],self.expected_row_counts()[tap_stream_id])

        # Verify that we have 'initial_full_table_complete' bookmark
        state = menagerie.get_state(conn_id)
        first_versions = {}

        for tap_stream_id in self.expected_check_streams():
            # assert that the state has an initial_full_table_complete == True
            self.assertTrue(state['bookmarks'][tap_stream_id]['initial_full_table_complete'])
            # assert that there is a version bookmark in state
            first_versions[tap_stream_id] = state['bookmarks'][tap_stream_id]['version']
            self.assertIsNotNone(first_versions[tap_stream_id])
            # Verify that we have a oplog_ts_time and oplog_ts_inc bookmark
            self.assertIsNotNone(state['bookmarks'][tap_stream_id]['oplog_ts_time'])
            self.assertIsNotNone(state['bookmarks'][tap_stream_id]['oplog_ts_inc'])

        changed_ids = set()
        with get_test_connection() as client:
            # Delete two documents for each collection

            changed_ids.add(client[db_name][self.coll_name_1].find({'int_field': 0})[0]['_id'])
            client[db_name][self.coll_name_1].delete_one({'int_field': 0})

            changed_ids.add(client[db_name][self.coll_name_1].find({'int_field': 1})[0]['_id'])
            client[db_name][self.coll_name_1].delete_one({'int_field': 1})

            changed_ids.add(client[db_name][self.coll_name_2].find({'int_field': 0})[0]['_id'])
            client[db_name][self.coll_name_2].delete_one({'int_field': 0})

            changed_ids.add(client[db_name][self.coll_name_2].find({'int_field': 1})[0]['_id'])
            client[db_name][self.coll_name_2].delete_one({'int_field': 1})

            # Update two documents for each collection
            changed_ids.add(client[db_name][self.coll_name_1].find({'int_field': 48})[0]['_id'])
            client[db_name][self.coll_name_1].update_one({'int_field': 48},{'$set': {'int_field': -1}})

            changed_ids.add(client[db_name][self.coll_name_1].find({'int_field': 49})[0]['_id'])
            client[db_name][self.coll_name_1].update_one({'int_field': 49},{'$set': {'int_field': -1}})

            changed_ids.add(client[db_name][self.coll_name_2].find({'int_field': 98})[0]['_id'])
            client[db_name][self.coll_name_2].update_one({'int_field': 98},{'$set': {'int_field': -1}})

            changed_ids.add(client[db_name][self.coll_name_2].find({'int_field': 99})[0]['_id'])
            client[db_name][self.coll_name_2].update_one({'int_field': 99},{'$set': {'int_field': -1}})

            # Insert two documents for each collection
            client[db_name][self.coll_name_1].insert_one({"int_field": 50, "string_field": random_string_generator()})
            changed_ids.add(client[db_name][self.coll_name_1].find({'int_field': 50})[0]['_id'])

            client[db_name][self.coll_name_1].insert_one({"int_field": 51, "string_field": random_string_generator()})
            changed_ids.add(client[db_name][self.coll_name_1].find({'int_field': 51})[0]['_id'])

            client[db_name][self.coll_name_2].insert_one({"int_field": 100, "string_field": random_string_generator()})
            changed_ids.add(client[db_name][self.coll_name_2].find({'int_field': 100})[0]['_id'])

            client[db_name][self.coll_name_2].insert_one({"int_field": 101, "string_field": random_string_generator()})
            changed_ids.add(client[db_name][self.coll_name_2].find({'int_field': 101})[0]['_id'])

        #  -------------------------------------------
        #  ----------- Subsequent Oplog Sync ---------
        #  -------------------------------------------

        # Run sync
        sync_job_name = runner.run_sync_mode(self, conn_id)

        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # verify the persisted schema was correct
        messages_by_stream = runner.get_records_from_target_output()
        records_by_stream = {}
        for stream_name in self.expected_sync_streams():
            records_by_stream[stream_name] = [x for x in messages_by_stream[stream_name]['messages']
                                              if x.get('action') == 'upsert']

        # assert that each of the streams that we synced are the ones that we expect to see
        record_count_by_stream = runner.examine_target_output_file(self,
                                                                   conn_id,
                                                                   self.expected_sync_streams(),
                                                                   self.expected_pks())

        # Verify that we got at least 6 records due to changes
        # (could be more due to overlap in gte oplog clause)
        for k,v in record_count_by_stream.items():
            self.assertGreaterEqual(v, 6)

        for stream in self.expected_sync_streams():
            # Verify that we got 2 records with _SDC_DELETED_AT
            self.assertEqual(2, len([x['data'] for x in records_by_stream[stream]
                                     if x['data'].get('_sdc_deleted_at')]))

        # Verify that the _id of the records sent are the same set as the
        # _ids of the documents changed
        actual_ids = {ObjectId(x['data']['_id']) for stream in self.expected_sync_streams()
                      for x in records_by_stream[stream]}

        self.assertEqual(changed_ids, actual_ids)



import bson
import os
import pdb
import random
import string
import unittest

from mongodb_common import drop_all_collections, get_test_connection, ensure_environment_variables_set
from tap_tester import connections, menagerie, runner


RECORD_COUNT = {}


def random_string_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))

def generate_simple_coll_docs(num_docs):
    docs = []
    for int_value in range(num_docs):
        docs.append({"int_field": int_value, "string_field": random_string_generator()})
    return docs

class MongoDBFullTable(unittest.TestCase):
    def setUp(self):
        ensure_environment_variables_set()

        with get_test_connection() as client:
            # drop all dbs/collections
            drop_all_collections(client)

            # simple_coll_1 has 50 documents
            client["simple_db"]["simple_coll_1"].insert_many(generate_simple_coll_docs(50))

            # create view on simple_coll_1
            client["simple_db"].command(bson.son.SON([("create", "simple_view_1"), ("viewOn", "simple_coll_1"), ("pipeline", [])]))

            # simple_coll_2 has 100 documents
            client["simple_db"]["simple_coll_2"].insert_many(generate_simple_coll_docs(100))

            # admin_coll_1 has 50 documents
            client["admin"]["admin_coll_1"].insert_many(generate_simple_coll_docs(50))

            # simple_coll_3 is an empty collection
            client["simple_db"].create_collection("simple_coll_3")

            # simple_coll_4 has documents with special chars and a lot of nesting
            client["simple_db"]["simple_coll_4"].insert_one({"hebrew_ישרא": "hebrew_ישרא"})
            client["simple_db"]["simple_coll_4"].insert_one({"hebrew_ישרא": 2})
            client["simple_db"]["simple_coll_4"].insert_one({"another_hebrew_ישראל": "another_hebrew_ישרא"})
            nested_doc = {"field0": {}}
            current_doc = nested_doc
            for i in range(1, 101):
                current_doc["field{}".format(i-1)]["field{}".format(i)] = {}
                current_doc = current_doc["field{}".format(i-1)]
            current_doc["field100"] = "some_value"
            client["simple_db"]["simple_coll_4"].insert_one(nested_doc)

            max_col_doc = {}
            for x in range(1600):
                max_col_doc['col_{}'.format(x)] = x
            client["simple_db"]["simple_coll_4"].insert_one(max_col_doc)




    def tap_stream_id_to_stream(self):
        return {
            'simple_db-simple_coll_1': 'simple_db_simple_coll_1',
            'simple_db-simple_coll_2': 'simple_db_simple_coll_2',
            'simple_db-simple_coll_3': 'simple_db_simple_coll_3',
            'simple_db-simple_coll_4': 'simple_db_simple_coll_4',
            'admin-admin_coll_1': 'admin_admin_coll_1'
        }

    def expected_check_streams(self):
        return {
            'simple_db-simple_coll_1',
            'simple_db-simple_coll_2',
            'simple_db-simple_coll_3',
            'simple_db-simple_coll_4',
            'admin-admin_coll_1'
        }

    def expected_pks(self):
        return {
            'simple_db_simple_coll_1': {'_id'},
            'simple_db_simple_coll_2': {'_id'},
            'simple_db_simple_coll_3': {'_id'},
            'simple_db_simple_coll_4': {'_id'},
            'admin_admin_coll_1': {'_id'}
        }

    def expected_row_counts(self):
        return {
            'simple_db_simple_coll_1': 50,
            'simple_db_simple_coll_2': 100,
            'simple_db_simple_coll_3': 0,
            'simple_db_simple_coll_4': 5,
            'admin_admin_coll_1': 50
        }

    def expected_sync_streams(self):
        return {
            'simple_db_simple_coll_1',
            'simple_db_simple_coll_2',
            'simple_db_simple_coll_3',
            'simple_db_simple_coll_4',
            'admin_admin_coll_1'
        }

    def name(self):
        return "tap_tester_mongodb_full_table"

    def tap_name(self):
        return "tap-mongodb"

    def get_type(self):
        return "platform.mongodb"

    def get_credentials(self):
        return {'password': os.getenv('TAP_MONGODB_PASSWORD')}

    def get_properties(self):
        return {'host' : os.getenv('TAP_MONGODB_HOST'),
                'port' : os.getenv('TAP_MONGODB_PORT'),
                'user' : os.getenv('TAP_MONGODB_USER'),
                'database' : os.getenv('TAP_MONGODB_DBNAME'),
                'include_schemas_in_destination_stream_name': 'true'
        }

    def test_run(self):

        conn_id = connections.ensure_connection(self)

        #  -------------------------------
        # -----------  Discovery ----------
        #  -------------------------------

        # run in discovery mode
        check_job_name = runner.run_check_mode(self, conn_id)

        # verify check exit codes
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        menagerie.verify_check_exit_status(self, exit_status, check_job_name)

        # verify the tap discovered the right streams
        found_catalogs = menagerie.get_catalogs(conn_id)

        # assert we find the correct streams
        self.assertEqual(self.expected_check_streams(),
                         {c['tap_stream_id'] for c in found_catalogs})

        #  -------------------------------------------
        #  ----------- First full Table Sync ---------
        #  -------------------------------------------
        # Select simple_coll_1 and simple_coll_2 streams and add replication method metadata
        for stream_catalog in found_catalogs:
            annotated_schema = menagerie.get_annotated_schema(conn_id, stream_catalog['stream_id'])
            additional_md = [{ "breadcrumb" : [], "metadata" : {'replication-method' : 'FULL_TABLE'}}]
            selected_metadata = connections.select_catalog_and_fields_via_metadata(conn_id,
                                                                                    stream_catalog,
                                                                                    annotated_schema,
                                                                                    additional_md)

        # run full table sync
        sync_job_name = runner.run_sync_mode(self, conn_id)

        # check exit status
        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # streams that we synced are the ones that we expect to see
        records_by_stream = runner.get_records_from_target_output()
        record_count_by_stream = runner.examine_target_output_file(self,
                                                                   conn_id,
                                                                   self.expected_sync_streams(),
                                                                   self.expected_pks())

        # assert that we get the correct number of records for each stream
        self.assertEqual(self.expected_row_counts(),record_count_by_stream)

        # assert that an activate_version_message is first and last message sent for each stream
        for stream_name in self.expected_sync_streams():
            self.assertEqual('activate_version',records_by_stream[stream_name]['messages'][0]['action'])
            self.assertEqual('activate_version',records_by_stream[stream_name]['messages'][-1]['action'])

        state = menagerie.get_state(conn_id)

        first_versions = {}

        for tap_stream_id in self.expected_check_streams():

            # state has an initial_full_table_complete == True
            self.assertTrue(state['bookmarks'][tap_stream_id]['initial_full_table_complete'])

            # there is a version bookmark in state
            first_versions[tap_stream_id] = state['bookmarks'][tap_stream_id]['version']
            self.assertIsNotNone(first_versions[tap_stream_id])

        #  -------------------------------------------
        #  ----------- Second full Table Sync ---------
        #  -------------------------------------------
        with get_test_connection() as client:
            # update existing documents in the collection to make sure we get the updates as well in the next sync
            doc_to_update = client["simple_db"]["simple_coll_1"].find_one()
            client["simple_db"]["simple_coll_1"].find_one_and_update({"_id": doc_to_update["_id"]}, {"$set": {"int_field": 999}})

            doc_to_update = client["simple_db"]["simple_coll_2"].find_one()
            client["simple_db"]["simple_coll_2"].find_one_and_update({"_id": doc_to_update["_id"]}, {"$set": {"int_field": 888}})

            doc_to_update = client["admin"]["admin_coll_1"].find_one()
            client["admin"]["admin_coll_1"].find_one_and_update({"_id": doc_to_update["_id"]}, {"$set": {"int_field": 777}})

            # add 2 rows and run full table again, make sure we get initial number + 2
            client["simple_db"]["simple_coll_1"].insert_many(generate_simple_coll_docs(2))

            client["simple_db"]["simple_coll_2"].insert_many(generate_simple_coll_docs(2))

            client["admin"]["admin_coll_1"].insert_many(generate_simple_coll_docs(2))

        sync_job_name = runner.run_sync_mode(self, conn_id)

        # check exit status
        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # verify the persisted schema was correct
        records_by_stream = runner.get_records_from_target_output()

        # assert that each of the streams that we synced are the ones that we expect to see
        record_count_by_stream = runner.examine_target_output_file(self,
                                                                   conn_id,
                                                                   self.expected_sync_streams(),
                                                                   self.expected_pks())

        state = menagerie.get_state(conn_id)

        # Verify  that menagerie state does not include a key for currently syncing
        self.assertIsNone(state['currently_syncing'])

        # Verify that menagerie state does not include a key for oplog based syncing
        self.assertNotIn('oplog', state)

        # assert that we have correct number of records (including the two new records and the update which is to be resynced)
        new_expected_row_counts = {k: v+2 for k, v in self.expected_row_counts().items() if k not in ['simple_db_simple_coll_3',
                                                                                                    'simple_db_simple_coll_4']}
        new_expected_row_counts['simple_db_simple_coll_3']=0
        new_expected_row_counts['simple_db_simple_coll_4']=5
        self.assertEqual(new_expected_row_counts, record_count_by_stream)

        # assert that we only have an ActivateVersionMessage as the last message and not the first
        for stream_name in self.expected_sync_streams():
            if len(records_by_stream[stream_name]['messages']) > 1:
                self.assertNotEqual('activate_version', records_by_stream[stream_name]['messages'][0]['action'], stream_name + "failed")
                self.assertEqual('upsert', records_by_stream[stream_name]['messages'][0]['action'], stream_name + "failed")
            self.assertEqual('activate_version', records_by_stream[stream_name]['messages'][-1]['action'], stream_name + "failed")

        second_versions = {}
        for tap_stream_id in self.expected_check_streams():
            found_stream = [c for c in found_catalogs if c['tap_stream_id'] == tap_stream_id][0]

            # state has an initial_full_table_complete == True
            self.assertTrue(state['bookmarks'][tap_stream_id]['initial_full_table_complete'])

            # version bookmark
            second_versions[tap_stream_id] = state['bookmarks'][tap_stream_id]['version']
            self.assertIsNotNone(second_versions[tap_stream_id])

            # version in this state is different than that of the previous state
            self.assertNotEqual(first_versions[tap_stream_id], second_versions[tap_stream_id])

            # version which is larger than the previous target version
            self.assertGreater(second_versions[tap_stream_id], first_versions[tap_stream_id])

            # verify that menagerie state does include the version which matches the target version
            self.assertEqual(records_by_stream[self.tap_stream_id_to_stream()[tap_stream_id]]['table_version'], second_versions[tap_stream_id])



import os
import pymongo
import random
import string
import unittest
from pymongo import ASCENDING

from mongodb_common import drop_all_collections, get_test_connection, ensure_environment_variables_set
from tap_tester import connections, menagerie, runner


RECORD_COUNT = {}


def random_string_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))

def generate_simple_coll_docs(num_docs):
    docs = []
    populated_string_fields = {f"string_field_{i}": random_string_generator() for i in range(1, 64)}
    for int_value in range(num_docs):
        docs.append({"int_field": int_value, **populated_string_fields})
    return docs

class MongoDBOplog(unittest.TestCase):
    def setUp(self):

        ensure_environment_variables_set()

        with get_test_connection() as client:
            ############# Drop all dbs/collections #############
            drop_all_collections(client)

            ############# Add simple collections #############
            # simple_coll_1 has 50 documents
            client["simple_db"]["simple_coll_1"].insert_many(generate_simple_coll_docs(50))

            # simple_coll_2 has 100 documents
            client["simple_db"]["simple_coll_2"].insert_many(generate_simple_coll_docs(100))

            for index in self.expected_string_fields():
                client["simple_db"]["simple_coll_1"].create_index(index)

            # # max 32 fields in a compound index (NO PLANS TO SUPPORT THIS IN THE TAP)
            # client["simple_db"]["simple_coll_1"].create_index([
            #     ('string_field', pymongo.ASCENDING), ('string_field_02', pymongo.ASCENDING),
            #     ('string_field_03', pymongo.ASCENDING), ('string_field_04', pymongo.ASCENDING),
            #     ('string_field_05', pymongo.ASCENDING), ('string_field_06', pymongo.ASCENDING),
            #     ('string_field_07', pymongo.ASCENDING), ('string_field_08', pymongo.ASCENDING),
            #     ('string_field_09', pymongo.ASCENDING), ('string_field_10', pymongo.ASCENDING),
            #     ('string_field_11', pymongo.ASCENDING), ('string_field_12', pymongo.ASCENDING),
            #     ('string_field_13', pymongo.ASCENDING), ('string_field_14', pymongo.ASCENDING),
            #     ('string_field_15', pymongo.ASCENDING), ('string_field_16', pymongo.ASCENDING),
            #     ('string_field_17', pymongo.ASCENDING), ('string_field_18', pymongo.ASCENDING),
            #     ('string_field_19', pymongo.ASCENDING), ('string_field_20', pymongo.ASCENDING),
            #     ('string_field_21', pymongo.ASCENDING), ('string_field_22', pymongo.ASCENDING),
            #     ('string_field_23', pymongo.ASCENDING), ('string_field_24', pymongo.ASCENDING),
            #     ('string_field_25', pymongo.ASCENDING), ('string_field_26', pymongo.ASCENDING),
            #     ('string_field_27', pymongo.ASCENDING), ('string_field_28', pymongo.ASCENDING),
            #     ('string_field_29', pymongo.ASCENDING), ('string_field_30', pymongo.ASCENDING),
            #     ('string_field_31', pymongo.ASCENDING), ('string_field_32', pymongo.ASCENDING)])

            self.index_info = client["simple_db"]["simple_coll_1"].index_information()

    def expected_check_streams(self):
        return {
            'simple_db-simple_coll_1',
            'simple_db-simple_coll_2',
        }

    def expected_pks(self):
        return {
            'simple_coll_1': {'_id'},
            'simple_coll_2': {'_id'},
        }

    def expected_row_counts(self):
        return {
            'simple_coll_1': 50,
            'simple_coll_2': 100,
        }

    def expected_sync_streams(self):
        return {
            'simple_coll_1',
            'simple_coll_2'
        }

    def name(self):
        return "tap_tester_mongodb_index"

    def tap_name(self):
        return "tap-mongodb"

    def get_type(self):
        return "platform.mongodb"

    def get_credentials(self):
        return {'password': os.getenv('TAP_MONGODB_PASSWORD')}

    def get_properties(self):
        return {'host' : os.getenv('TAP_MONGODB_HOST'),
                'port' : os.getenv('TAP_MONGODB_PORT'),
                'user' : os.getenv('TAP_MONGODB_USER'),
                'database' : os.getenv('TAP_MONGODB_DBNAME')
        }

    def expected_string_fields(self):
        # Max index count = 64.  63 strings + '_id'
        return {f"string_field_{i}" for i in range(1, 64)}


    def test_run(self):

        conn_id = connections.ensure_connection(self)

        #  -----------------------------------
        #  -----------  Discovery ------------
        #  -----------------------------------

        # run in discovery mode
        check_job_name = runner.run_check_mode(self, conn_id)

        # verify check  exit codes
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        menagerie.verify_check_exit_status(self, exit_status, check_job_name)

        # verify the tap discovered the right streams
        found_catalogs = menagerie.get_catalogs(conn_id)

        # assert we find the correct streams
        self.assertEqual(self.expected_check_streams(),
                         {c['tap_stream_id'] for c in found_catalogs})

        for tap_stream_id in self.expected_check_streams():
            found_stream = [c for c in found_catalogs if c['tap_stream_id'] == tap_stream_id][0]

            # assert that the pks are correct
            self.assertEqual(self.expected_pks()[found_stream['stream_name']],
                             set(found_stream.get('metadata', {}).get('table-key-properties')))

            # assert that the row counts are correct
            self.assertEqual(self.expected_row_counts()[found_stream['stream_name']],
                             found_stream.get('metadata', {}).get('row-count'))

        # no plans for tap to support compound index, may not appear in valid-replication-keys list
        discovered_replication_keys = found_catalogs[0]['metadata']['valid-replication-keys']
        for field in self.expected_string_fields():
            self.assertIn(field, discovered_replication_keys)
        self.assertIn('_id', discovered_replication_keys)
        self.assertEqual(64, len(discovered_replication_keys))



import os
import random
import string
import unittest
from bson import ObjectId
from pymongo.errors import BulkWriteError, WriteError

from mongodb_common import drop_all_collections, ensure_environment_variables_set, \
    get_test_connection
from tap_tester import connections, menagerie, runner, LOGGER


RECORD_COUNT = {}


def random_string_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))

def generate_simple_coll_docs(num_docs):
    docs = []
    for int_value in range(num_docs):
        # TDL-20088. Version 4.2 does not support leading "." or leading "$" in
        # field names. Also unsupported: use of field names with "." in find() queries
        docs.append({"int.field": int_value, "string$field": random_string_generator()})
    return docs

def generate_leading_coll_docs(num_docs):
    docs = []
    for int_value in range(num_docs):
        # TDL-20088. Versions 4.2 and 4.4 do not support leading "." or leading "$" in
        # field names. Also unsupported: use of field names with "." in find() queries
        docs.append({".int_field": int_value, "$string_field": random_string_generator()})
    return docs

class MongoDBFieldNameRestrictions(unittest.TestCase):
    ''' Test edge case field name restrictions per the documentation (use of '.' and '$')
        Reference https://jira.talendforge.org/browse/TDL-18990 for details  '''

    def setUp(self):

        ensure_environment_variables_set()

        with get_test_connection() as client:

            self.db_version = client.server_info()['version']
            LOGGER.info("MongoDB version {}".format(self.db_version))

            ############# Drop all dbs/collections #############
            drop_all_collections(client)

            ############# Add simple collections #############
            # simple_coll_1 has 50 documents
            client["simple_db"]["simple_coll_1"].insert_many(generate_simple_coll_docs(50))

            # simple_coll_2 has 100 documents
            client["simple_db"]["simple_coll_2"].insert_many(generate_simple_coll_docs(100))

            if self.db_version == '4.4.6':  # TODO add 4.2 if image is available to test against
                with self.assertRaises(BulkWriteError) as e:
                    # simple_coll_3 attempts to insert 10 documents
                    client["simple_db"]["simple_coll_3"].insert_many(generate_leading_coll_docs(10))

                e_error = "Document can't have $ prefixed field names"
                self.assertIn(e_error, str(e.exception.details.values()))

            else:
                # simple_coll_3 has 10 documents
                client["simple_db"]["simple_coll_3"].insert_many(generate_leading_coll_docs(10))

    def expected_check_streams(self):
        if self.db_version == '4.4.6':
            return {
                'simple_db-simple_coll_1',
                'simple_db-simple_coll_2',
            }
        return {
            'simple_db-simple_coll_1',
            'simple_db-simple_coll_2',
            'simple_db-simple_coll_3',
        }

    def expected_pks(self):
        if self.db_version == '4.4.6':
            return {
                'simple_coll_1': {'_id'},
                'simple_coll_2': {'_id'},
            }
        return {
            'simple_coll_1': {'_id'},
            'simple_coll_2': {'_id'},
            'simple_coll_3': {'_id'},
        }

    def expected_row_counts(self):
        if self.db_version == '4.4.6':
            return {
                'simple_coll_1': 50,
                'simple_coll_2': 100,
            }
        return {
            'simple_coll_1': 50,
            'simple_coll_2': 100,
            'simple_coll_3': 10,
        }

    def expected_sync_streams(self):
        if self.db_version == '4.4.6':
            return {
                'simple_coll_1',
                'simple_coll_2',
            }
        return {
            'simple_coll_1',
            'simple_coll_2',
            'simple_coll_3',
        }

    def name(self):
        return "tap_tester_mongodb_fname_restrict"

    def tap_name(self):
        return "tap-mongodb"

    def get_type(self):
        return "platform.mongodb"

    def get_credentials(self):
        return {'password': os.getenv('TAP_MONGODB_PASSWORD')}

    def get_properties(self):
        return {'host' : os.getenv('TAP_MONGODB_HOST'),
                'port' : os.getenv('TAP_MONGODB_PORT'),
                'user' : os.getenv('TAP_MONGODB_USER'),
                'database' : os.getenv('TAP_MONGODB_DBNAME')
        }

    def test_run(self):

        conn_id = connections.ensure_connection(self)

        #  ---------------------------------
        #  -----------  Discovery ----------
        #  ---------------------------------

        # run in discovery mode
        check_job_name = runner.run_check_mode(self, conn_id)

        # verify check  exit codes
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        menagerie.verify_check_exit_status(self, exit_status, check_job_name)

        # verify the tap discovered the right streams
        found_catalogs = menagerie.get_catalogs(conn_id)

        # assert we find the correct streams
        self.assertEqual(self.expected_check_streams(),
                         {c['tap_stream_id'] for c in found_catalogs})


        for tap_stream_id in self.expected_check_streams():
            found_stream = [c for c in found_catalogs if c['tap_stream_id'] == tap_stream_id][0]

            # assert that the pks are correct
            self.assertEqual(self.expected_pks()[found_stream['stream_name']],
                             set(found_stream.get('metadata', {}).get('table-key-properties')))

            # assert that the row counts are correct
            self.assertEqual(self.expected_row_counts()[found_stream['stream_name']],
                             found_stream.get('metadata', {}).get('row-count'))

        #  ----------------------------------------
        #  ----------- Initial Full Table ---------
        #  ----------------------------------------

        # Select simple_coll_1 and simple_coll_2 streams and add replication method metadata
        for stream_catalog in found_catalogs:
            annotated_schema = menagerie.get_annotated_schema(conn_id, stream_catalog['stream_id'])
            additional_md = [{"breadcrumb" : [], "metadata" : {'replication-method' : 'LOG_BASED'}}]
            selected_metadata = connections.select_catalog_and_fields_via_metadata(conn_id,
                                                                                   stream_catalog,
                                                                                   annotated_schema,
                                                                                   additional_md)

        # Run sync
        sync_job_name = runner.run_sync_mode(self, conn_id)

        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # verify the persisted schema was correct
        records_by_stream = runner.get_records_from_target_output()

        # assert that each of the streams that we synced are the ones that we expect to see
        record_count_by_stream = runner.examine_target_output_file(self,
                                                                   conn_id,
                                                                   self.expected_sync_streams(),
                                                                   self.expected_pks())

        # Verify that the full table was synced
        for tap_stream_id in self.expected_sync_streams():
            self.assertGreaterEqual(record_count_by_stream[tap_stream_id],
                                    self.expected_row_counts()[tap_stream_id])

        # Verify that we have 'initial_full_table_complete' bookmark
        state = menagerie.get_state(conn_id)
        first_versions = {}

        for tap_stream_id in self.expected_check_streams():
            # assert that the state has an initial_full_table_complete == True
            self.assertTrue(state['bookmarks'][tap_stream_id]['initial_full_table_complete'])
            # assert that there is a version bookmark in state
            first_versions[tap_stream_id] = state['bookmarks'][tap_stream_id]['version']
            self.assertIsNotNone(first_versions[tap_stream_id])
            # Verify that we have a oplog_ts_time and oplog_ts_inc bookmark
            self.assertIsNotNone(state['bookmarks'][tap_stream_id]['oplog_ts_time'])
            self.assertIsNotNone(state['bookmarks'][tap_stream_id]['oplog_ts_inc'])

        changed_ids = set()
        failed_update_ids = set()
        with get_test_connection() as client:
            # Delete two documents for each collection
            object_id = client['simple_db']['simple_coll_1'].find()[0]['_id']
            changed_ids.add(object_id)
            client["simple_db"]["simple_coll_1"].delete_one({'_id': object_id})

            object_id = client['simple_db']['simple_coll_1'].find()[1]['_id']
            changed_ids.add(object_id)
            client["simple_db"]["simple_coll_1"].delete_one({'_id': object_id})

            object_id = client['simple_db']['simple_coll_2'].find()[0]['_id']
            changed_ids.add(object_id)
            client["simple_db"]["simple_coll_2"].delete_one({'_id': object_id})

            object_id = client['simple_db']['simple_coll_2'].find()[1]['_id']
            changed_ids.add(object_id)
            client["simple_db"]["simple_coll_2"].delete_one({'_id': object_id})

            if self.db_version != '4.4.6':
                object_id = client['simple_db']['simple_coll_3'].find()[0]['_id']
                changed_ids.add(object_id)
                client["simple_db"]["simple_coll_3"].delete_one({'_id': object_id})

                object_id = client['simple_db']['simple_coll_3'].find()[1]['_id']
                changed_ids.add(object_id)
                client["simple_db"]["simple_coll_3"].delete_one({'_id': object_id})

            # Update two documents for each collection

            # curor objects do not support negative indicies so set indicies for last two records
            num_records = len(list(client['simple_db']['simple_coll_1'].find()))
            last_index = num_records - 1
            sec_last_index = num_records -2

            # Not supported in mongodb 4.2 TDL-20088
            # int.field 48
            # object_id = client['simple_db']['simple_coll_1'].find()[sec_last_index]['_id']
            # changed_ids.add(object_id)
            # client["simple_db"]["simple_coll_1"].update_one(
            #     {'_id': object_id},{'$set': {'int.field': -1}})

            # int.field 49
            object_id = client['simple_db']['simple_coll_1'].find()[last_index]['_id']
            changed_ids.add(object_id)
            client["simple_db"]["simple_coll_1"].update_one(
                {'_id': object_id},{'$set': {'string$field': "Updated_1"}})

            num_records = len(list(client['simple_db']['simple_coll_2'].find()))
            last_index = num_records - 1
            sec_last_index = num_records - 2

            # Not supported in mongodb 4.2 TDL-20088
            # int.field 98
            # object_id = client['simple_db']['simple_coll_2'].find()[sec_last_index]['_id']
            # changed_ids.add(object_id)
            # client["simple_db"]["simple_coll_2"].update_one(
            #     {'_id': object_id},{'$set': {'int.field': -1}})

            # int.field 99
            object_id = client['simple_db']['simple_coll_2'].find()[last_index]['_id']
            changed_ids.add(object_id)
            client["simple_db"]["simple_coll_2"].update_one(
                {'_id': object_id},{'$set': {'string$field': "Updated_2"}})

            if self.db_version != '4.4.6':
                num_records = len(list(client['simple_db']['simple_coll_3'].find()))
                last_index = num_records - 1
                sec_last_index = num_records - 2

                object_id = client['simple_db']['simple_coll_3'].find()[sec_last_index]['_id']
                changed_ids.add(object_id)  # this results in empty field and does not replicate
                failed_update_ids.add(object_id)
                with self.assertRaises(WriteError) as e1:
                    client["simple_db"]["simple_coll_3"].update_one(
                        {'_id': object_id},{'$set': {'.int_field': -1}})

                e1_error = ".int_field' contains an empty field name, which is not allowed"
                self.assertIn(e1_error, str(e1.exception.details.values()))

                object_id = client['simple_db']['simple_coll_3'].find()[last_index]['_id']
                changed_ids.add(object_id)
                failed_update_ids.add(object_id)
                with self.assertRaises(WriteError) as e2:
                    client["simple_db"]["simple_coll_3"].update_one(
                        {'_id': object_id},{'$set': {'$string_field': "Updated_3"}})

                e2_error = ( "The dollar ($) prefixed field '$string_field' in '$string_field' is not "
                             "allowed in the context of an update's replacement document." )
                self.assertIn(e2_error, str(e2.exception.details.values()))

            # Insert two documents for each collection
            last_index = len(list(client['simple_db']['simple_coll_1'].find()))
            client["simple_db"]["simple_coll_1"].insert_one(
                {"int.field": 50, "string$field": random_string_generator()})
            object_id = client["simple_db"]["simple_coll_1"].find()[last_index]['_id']
            changed_ids.add(object_id)

            last_index += 1 # inserting a new item
            client["simple_db"]["simple_coll_1"].insert_one(
                {"int.field": 51, "string$field": random_string_generator()})
            object_id = client["simple_db"]["simple_coll_1"].find()[last_index]['_id']
            changed_ids.add(object_id)

            last_index = len(list(client['simple_db']['simple_coll_2'].find()))
            client["simple_db"]["simple_coll_2"].insert_one(
                {"int.field": 100, "string$field": random_string_generator()})
            object_id = client["simple_db"]["simple_coll_2"].find()[last_index]['_id']
            changed_ids.add(object_id)

            last_index += 1
            client["simple_db"]["simple_coll_2"].insert_one(
                {"int.field": 101, "string$field": random_string_generator()})
            object_id = client["simple_db"]["simple_coll_2"].find()[last_index]['_id']
            changed_ids.add(object_id)

            if self.db_version != '4.4.6':
                last_index = len(list(client['simple_db']['simple_coll_3'].find()))
                client["simple_db"]["simple_coll_3"].insert_one(
                    {".int_field": 300, "$string_field": random_string_generator()})
                object_id = client["simple_db"]["simple_coll_3"].find()[last_index]['_id']
                changed_ids.add(object_id)

                last_index += 1
                client["simple_db"]["simple_coll_3"].insert_one(
                    {".int_field": 301, "$string_field": random_string_generator()})
                object_id = client["simple_db"]["simple_coll_3"].find()[last_index]['_id']
                changed_ids.add(object_id)

        #  -------------------------------------------
        #  ----------- Subsequent Oplog Sync ---------
        #  -------------------------------------------

        # Run sync
        sync_job_name = runner.run_sync_mode(self, conn_id)

        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        # verify the persisted schema was correct
        messages_by_stream = runner.get_records_from_target_output()
        records_by_stream = {}
        for stream_name in self.expected_sync_streams():
            records_by_stream[stream_name] = [x for x in messages_by_stream[stream_name]['messages']
                                              if x.get('action') == 'upsert']

        # assert that each of the streams that we synced are the ones that we expect to see
        record_count_by_stream = runner.examine_target_output_file(self,
                                                                   conn_id,
                                                                   self.expected_sync_streams(),
                                                                   self.expected_pks())

        # Verify that we got at least 5 records due to changes
        # (could be more due to overlap in gte oplog clause)
        for k,v in record_count_by_stream.items():
            # updates for fields with leading . or $ characters may not replicate
            if k == 'simple_coll_3':
                self.assertGreaterEqual(v, 4)
            else:
                self.assertGreaterEqual(v, 5)

        for stream in self.expected_sync_streams():
            # Verify that we got 2 records with _SDC_DELETED_AT
            self.assertEqual(2, len([x['data'] for x in records_by_stream[stream]
                                     if x['data'].get('_sdc_deleted_at')]))

        # Verify that the _id of the records sent are the same set as the
        # _ids of the documents changed
        actual_ids = {ObjectId(x['data']['_id']) for stream in self.expected_sync_streams()
                      for x in records_by_stream[stream]}

        adjusted_changed_ids = changed_ids.difference(failed_update_ids)
        adjusted_actual_ids = actual_ids.difference(failed_update_ids)
        self.assertEqual(adjusted_changed_ids, adjusted_actual_ids)

        with get_test_connection() as client:
            # Verify the updated record values in the db match the tap output file
            found_id_db = client['simple_db']['simple_coll_1'].find({'string$field': "Updated_1"})[0]['_id']
            found_id_tap = [ x['data']['_id'] for x in records_by_stream['simple_coll_1']
                                    if x['data'].get('string$field') == 'Updated_1' ]
            self.assertEqual(str(found_id_db), found_id_tap[0])

            found_id_db = client['simple_db']['simple_coll_2'].find({'string$field': "Updated_2"})[0]['_id']
            found_id_tap = [ x['data']['_id'] for x in records_by_stream['simple_coll_2']
                                    if x['data'].get('string$field') == 'Updated_2' ]
            self.assertEqual(str(found_id_db), found_id_tap[0])

            # simple_coll_3 updates are expected to fail so there is nothing to verify

            # Not supported in mongodb 4.2 TDL-20088. Verify updated record values in db match tap output file
            # found_int = -1 # one document in each colleciton was updated to have an int value of -1
            # found_ints_1 = [x['data']['int.field'] for x in records_by_stream['simple_coll_1'] if x['data'].get('int.field')]
            # found_ints_2 = [x['data']['int.field'] for x in records_by_stream['simple_coll_2'] if x['data'].get('int.field')]
            # self.assertIn(found_int, found_ints_1)
            # self.assertIn(found_int, found_ints_2)
